@@ -3,18 +3,22 @@
  *
  * Copyright (C) 2004,2005  Theodore Ts'o <tytso@mit.edu>
  *
- * This file can be redistributed under the terms of the GNU Library General
- * Public License
+ * %Begin-Header%
+ * This file may be redistributed under the terms of the GNU Library
+ * General Public License, version 2.
+ * %End-Header%
  */
 
+#include "config.h"
 #include "e2p.h"
 #include <string.h>
+#include <stdlib.h>
 
-const char *os_tab[] =
-	{ "Linux", 
-	  "Hurd", 
-	  "Masix", 
-	  "FreeBSD", 
+static const char *os_tab[] =
+	{ "Linux",
+	  "Hurd",
+	  "Masix",
+	  "FreeBSD",
 	  "Lites",
 	  0 };
 
@@ -32,7 +36,8 @@ char *e2p_os2string(int os_type)
 		os = "(unknown os)";
 
         ret = malloc(strlen(os)+1);
-        strcpy(ret, os);
+	if (ret)
+		strcpy(ret, os);
         return ret;
 }
 
@@ -61,6 +66,7 @@ int main(int argc, char **argv)
 		s = e2p_os2string(i);
 		os = e2p_string2os(s);
 		printf("%d: %s (%d)\n", i, s, os);
+		free(s);
 		if (i != os) {
 			fprintf(stderr, "Failure!\n");
 			exit(1);
